@@ -2,6 +2,8 @@
 #include "draw.h"
 #include "log.h"
 
+#include <limits.h>
+
 void
 i_mmarray_cr(i_mmarray *ar,int l) {
   int i;
@@ -44,6 +46,55 @@ i_mmarray_render(i_img *im,i_mmarray *ar,i_color *val) {
   for(i=0;i<ar->lines;i++) if (ar->data[i].max!=-1) for(x=ar->data[i].min;x<ar->data[i].max;x++) i_ppix(im,x,i,val);
 }
 
+void
+i_mmarray_render_fill(i_img *im,i_mmarray *ar,i_fill_t *fill) {
+  int x, w, y;
+  if (im->bits == i_8_bits && fill->fill_with_color) {
+    i_color *line = mymalloc(sizeof(i_color) * im->xsize);
+    i_color *work = NULL;
+    if (fill->combine)
+      work = mymalloc(sizeof(i_color) * im->xsize);
+    for(y=0;y<ar->lines;y++) {
+      if (ar->data[y].max!=-1) {
+        x = ar->data[y].min;
+        w = ar->data[y].max-ar->data[y].min;
+
+        if (fill->combine) 
+          i_glin(im, x, x+w, y, line);
+        
+        (fill->fill_with_color)(fill, x, y, w, im->channels, line, work);
+        i_plin(im, x, x+w, y, line);
+      }
+    }
+  
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
+  else {
+    i_fcolor *line = mymalloc(sizeof(i_fcolor) * im->xsize);
+    i_fcolor *work = NULL;
+    if (fill->combinef)
+      work = mymalloc(sizeof(i_fcolor) * im->xsize);
+    for(y=0;y<ar->lines;y++) {
+      if (ar->data[y].max!=-1) {
+        x = ar->data[y].min;
+        w = ar->data[y].max-ar->data[y].min;
+
+        if (fill->combinef) 
+          i_glinf(im, x, x+w, y, line);
+        
+        (fill->fill_with_fcolor)(fill, x, y, w, im->channels, line, work);
+        i_plinf(im, x, x+w, y, line);
+      }
+    }
+  
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
+}
+
 
 static
 void
@@ -106,14 +157,192 @@ i_arc(i_img *im,int x,int y,float rad,float d1,float d2,i_color *val) {
   y1=(int)(y+0.5+rad*sin(d2*PI/180.0));
 
   for(f=d1;f<=d2;f+=0.01) i_mmarray_add(&dot,(int)(x+0.5+rad*cos(f*PI/180.0)),(int)(y+0.5+rad*sin(f*PI/180.0)));
-
+  
   /*  printf("x1: %d.\ny1: %d.\n",x1,y1); */
   i_arcdraw(x, y, x1, y1, &dot);
 
   /*  dot.info(); */
   i_mmarray_render(im,&dot,val);
+  i_mmarray_dst(&dot);
 }
 
+void
+i_arc_cfill(i_img *im,int x,int y,float rad,float d1,float d2,i_fill_t *fill) {
+  i_mmarray dot;
+  float f,fx,fy;
+  int x1,y1;
+
+  mm_log((1,"i_arc_cfill(im* 0x%x,x %d,y %d,rad %.2f,d1 %.2f,d2 %.2f,fill 0x%x)\n",im,x,y,rad,d1,d2,fill));
+
+  i_mmarray_cr(&dot,im->ysize);
+
+  x1=(int)(x+0.5+rad*cos(d1*PI/180.0));
+  y1=(int)(y+0.5+rad*sin(d1*PI/180.0));
+  fx=(float)x1; fy=(float)y1;
+
+  /*  printf("x1: %d.\ny1: %d.\n",x1,y1); */
+  i_arcdraw(x, y, x1, y1, &dot);
+
+  x1=(int)(x+0.5+rad*cos(d2*PI/180.0));
+  y1=(int)(y+0.5+rad*sin(d2*PI/180.0));
+
+  for(f=d1;f<=d2;f+=0.01) i_mmarray_add(&dot,(int)(x+0.5+rad*cos(f*PI/180.0)),(int)(y+0.5+rad*sin(f*PI/180.0)));
+  
+  /*  printf("x1: %d.\ny1: %d.\n",x1,y1); */
+  i_arcdraw(x, y, x1, y1, &dot);
+
+  /*  dot.info(); */
+  i_mmarray_render_fill(im,&dot,fill);
+  i_mmarray_dst(&dot);
+}
+
+
+
+/* Temporary AA HACK */
+
+
+typedef int frac;
+static  frac float_to_frac(float x) { return (frac)(0.5+x*16.0); }
+static   int frac_sub     (frac x)  { return (x%16); }
+static   int frac_int     (frac x)  { return (x/16); }
+static float frac_to_float(float x) { return (float)x/16.0; }
+
+static 
+void
+polar_to_plane(float cx, float cy, float angle, float radius, frac *x, frac *y) {
+  *x = float_to_frac(cx+radius*cos(angle));
+  *y = float_to_frac(cy+radius*sin(angle));
+}
+
+static
+void
+order_pair(frac *x, frac *y) {
+  frac t = *x;
+  if (t>*y) {
+    *x = *y;
+    *y = t;
+  }
+}
+
+
+
+
+static
+void
+make_minmax_list(i_mmarray *dot, float x, float y, float radius) {
+  float angle = 0.0;
+  float astep = radius>0.1 ? .5/radius : 10;
+  frac cx, cy, lx, ly, sx, sy;
+
+  mm_log((1, "make_minmax_list(dot %p, x %.2f, y %.2f, radius %.2f)\n", dot, x, y, radius));
+
+  polar_to_plane(x, y, angle, radius, &sx, &sy);
+  
+  for(angle = 0.0; angle<361; angle +=astep) {
+    float alpha;
+    lx = sx; ly = sy;
+    polar_to_plane(x, y, angle, radius, &cx, &cy);
+    sx = cx; sy = cy;
+
+    if (fabs(cx-lx) > fabs(cy-ly)) {
+      int ccx, ccy;
+      if (lx>cx) { 
+	ccx = lx; lx = cx; cx = ccx; 
+	ccy = ly; ly = cy; cy = ccy; 
+      }
+
+      for(ccx=lx; ccx<=cx; ccx++) {
+	ccy = ly + ((cy-ly)*(ccx-lx))/(cx-lx);
+	i_mmarray_add(dot, ccx, ccy);
+      }
+    } else {
+      int ccx, ccy;
+
+      if (ly>cy) { 
+	ccy = ly; ly = cy; cy = ccy; 
+	ccx = lx; lx = cx; cx = ccx; 
+      }
+      
+      for(ccy=ly; ccy<=cy; ccy++) {
+	if (cy-ly) ccx = lx + ((cx-lx)*(ccy-ly))/(cy-ly); else ccx = lx;
+	i_mmarray_add(dot, ccx, ccy);
+      }
+    }
+  }
+}
+
+/* Get the number of subpixels covered */
+
+static
+int
+i_pixel_coverage(i_mmarray *dot, int x, int y) {
+  frac minx = x*16;
+  frac maxx = minx+15;
+  frac cy;
+  int cnt = 0;
+  
+  for(cy=y*16; cy<(y+1)*16; cy++) {
+    frac tmin = dot->data[cy].min;
+    frac tmax = dot->data[cy].max;
+
+    if (tmax == -1 || tmin > maxx || tmax < minx) continue;
+    
+    if (tmin < minx) tmin = minx;
+    if (tmax > maxx) tmax = maxx;
+    
+    cnt+=1+tmax-tmin;
+  }
+  return cnt;
+}
+
+void
+i_circle_aa(i_img *im, float x, float y, float rad, i_color *val) {
+  i_mmarray dot;
+  i_color temp;
+  int ly;
+
+  mm_log((1, "i_circle_aa(im %p, x %d, y %d, rad %.2f, val %p)\n", im, x, y, rad, val));
+
+  i_mmarray_cr(&dot,16*im->ysize);
+  make_minmax_list(&dot, x, y, rad);
+
+  for(ly = 0; ly<im->ysize; ly++) {
+    int ix, cy, cnt = 0, minx = INT_MAX, maxx = INT_MIN;
+
+    /* Find the left/rightmost set subpixels */
+    for(cy = 0; cy<16; cy++) {
+      frac tmin = dot.data[ly*16+cy].min;
+      frac tmax = dot.data[ly*16+cy].max;
+      if (tmax == -1) continue;
+
+      if (minx > tmin) minx = tmin;
+      if (maxx < tmax) maxx = tmax;
+    }
+
+    if (maxx == INT_MIN) continue; /* no work to be done for this row of pixels */
+
+    minx /= 16;
+    maxx /= 16;
+    for(ix=minx; ix<=maxx; ix++) {
+      int cnt = i_pixel_coverage(&dot, ix, ly);
+      if (cnt>255) cnt = 255;
+      if (cnt) { /* should never be true */
+	int ch;
+	float ratio = (float)cnt/255.0;
+	i_gpix(im, ix, ly, &temp);
+	for(ch=0;ch<im->channels; ch++) temp.channel[ch] = (unsigned char)((float)val->channel[ch]*ratio + (float)temp.channel[ch]*(1.0-ratio));
+	i_ppix(im, ix, ly, &temp);
+      }
+    }
+  }
+  i_mmarray_dst(&dot);
+}
+
+
+
+
+
+
 void
 i_box(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
   int x,y;
@@ -135,6 +364,46 @@ i_box_filled(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
   for(x=x1;x<x2+1;x++) for (y=y1;y<y2+1;y++) i_ppix(im,x,y,val);
 }
 
+void
+i_box_cfill(i_img *im,int x1,int y1,int x2,int y2,i_fill_t *fill) {
+  mm_log((1,"i_box_cfill(im* 0x%x,x1 %d,y1 %d,x2 %d,y2 %d,fill 0x%x)\n",im,x1,y1,x2,y2,fill));
+
+  ++x2;
+  if (im->bits == i_8_bits && fill->fill_with_color) {
+    i_color *line = mymalloc(sizeof(i_color) * (x2 - x1));
+    i_color *work = NULL;
+    if (fill->combine)
+      work = mymalloc(sizeof(i_color) * (x2-x1));
+    while (y1 <= y2) {
+      if (fill->combine)
+        i_glin(im, x1, x2, y1, line);
+
+      (fill->fill_with_color)(fill, x1, y1, x2-x1, im->channels, line, work);
+      i_plin(im, x1, x2, y1, line);
+      ++y1;
+    }
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
+  else {
+    i_fcolor *line = mymalloc(sizeof(i_fcolor) * (x2 - x1));
+    i_fcolor *work;
+    work = mymalloc(sizeof(i_fcolor) * (x2 - x1));
+
+    while (y1 <= y2) {
+      if (fill->combinef)
+        i_glinf(im, x1, x2, y1, line);
+
+      (fill->fill_with_fcolor)(fill, x1, y1, x2-x1, im->channels, line, work);
+      i_plinf(im, x1, x2, y1, line);
+      ++y1;
+    }
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
+}
 
 void
 i_draw(i_img *im,int x1,int y1,int x2,int y2,i_color *val) {
@@ -293,21 +562,6 @@ i_bezier_multi(i_img *im,int l,double *x,double *y,i_color *val) {
 
 */
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
 /* This should be moved into a seperate file? */
 
 /* This is the truncation used:
@@ -672,13 +926,13 @@ static
 struct stack_element*
 crdata(int left,int right,int dadl,int dadr,int y, int dir) {
   struct stack_element *ste;
-  ste=(struct stack_element*)mymalloc(sizeof(struct stack_element));
-  ste->myLx=left;
-  ste->myRx=right;
-  ste->dadLx=dadl;
-  ste->dadRx=dadr;
-  ste->myY=y;
-  ste->myDirection=dir;
+  ste              = mymalloc(sizeof(struct stack_element));
+  ste->myLx        = left;
+  ste->myRx        = right;
+  ste->dadLx       = dadl;
+  ste->dadRx       = dadr;
+  ste->myY         = y;
+  ste->myDirection = dir;
   return ste;
 }
 
@@ -718,14 +972,14 @@ i_rspan(i_img *im,int seedx,int seedy,i_color *val) {
 
 /* Macro to create a link and push on to the list */
 
-#define ST_PUSH(left,right,dadl,dadr,y,dir) { struct stack_element *s=crdata(left,right,dadl,dadr,y,dir); llist_push(st,&s);}
+#define ST_PUSH(left,right,dadl,dadr,y,dir) { struct stack_element *s = crdata(left,right,dadl,dadr,y,dir); llist_push(st,&s); }
 
 /* pops the shadow on TOS into local variables lx,rx,y,direction,dadLx and dadRx */
 /* No overflow check! */
  
-#define ST_POP() { struct stack_element *s; llist_pop(st,&s); lx=s->myLx; rx=s->myRx; dadLx=s->dadLx; dadRx=s->dadRx; y=s->myY; direction=s->myDirection; myfree(s); }
+#define ST_POP() { struct stack_element *s; llist_pop(st,&s); lx = s->myLx; rx = s->myRx; dadLx = s->dadLx; dadRx = s->dadRx; y = s->myY; direction= s->myDirection; myfree(s); }
 
-#define ST_STACK(dir,dadLx,dadRx,lx,rx,y) { int pushrx=rx+1; int pushlx=lx-1; ST_PUSH(lx,rx,pushlx,pushrx,y+dir,dir); if (rx > dadRx) ST_PUSH(dadRx+1,rx,pushlx,pushrx,y-dir,-dir); if (lx < dadLx) ST_PUSH(lx,dadLx-1,pushlx,pushrx,y-dir,-dir); }
+#define ST_STACK(dir,dadLx,dadRx,lx,rx,y) { int pushrx = rx+1; int pushlx = lx-1; ST_PUSH(lx,rx,pushlx,pushrx,y+dir,dir); if (rx > dadRx)                                      ST_PUSH(dadRx+1,rx,pushlx,pushrx,y-dir,-dir); if (lx < dadLx) ST_PUSH(lx,dadLx-1,pushlx,pushrx,y-dir,-dir); }
 
 #define SET(x,y) btm_set(btm,x,y);
 
@@ -752,6 +1006,134 @@ i_flood_fill(i_img *im,int seedx,int seedy,i_color *dcol) {
   int channels,xsize,ysize;
   i_color cval,val;
 
+  channels = im->channels;
+  xsize    = im->xsize;
+  ysize    = im->ysize;
+
+  btm = btm_new(xsize,ysize);
+  st = llist_new(100,sizeof(struct stack_element*));
+
+  /* Get the reference color */
+  i_gpix(im,seedx,seedy,&val);
+
+  /* Find the starting span and fill it */
+  lx = i_lspan(im,seedx,seedy,&val);
+  rx = i_rspan(im,seedx,seedy,&val);
+  
+  /* printf("span: %d %d \n",lx,rx); */
+
+  for(x=lx; x<=rx; x++) SET(x,seedy);
+
+  ST_PUSH(lx, rx, lx, rx, seedy+1, 1);
+  ST_PUSH(lx, rx, lx, rx, seedy-1,-1);
+
+  while(st->count) {
+    ST_POP();
+    
+    if (y<0 || y>ysize-1) continue;
+
+    if (bymin > y) bymin=y; /* in the worst case an extra line */
+    if (bymax < y) bymax=y; 
+
+    /*     printf("start of scan - on stack : %d \n",st->count); */
+
+    
+    /*     printf("lx=%d rx=%d dadLx=%d dadRx=%d y=%d direction=%d\n",lx,rx,dadLx,dadRx,y,direction); */
+    
+    /*
+    printf(" ");
+    for(tx=0;tx<xsize;tx++) printf("%d",tx%10);
+    printf("\n");
+    for(ty=0;ty<ysize;ty++) {
+      printf("%d",ty%10);
+      for(tx=0;tx<xsize;tx++) printf("%d",!!btm_test(btm,tx,ty));
+      printf("\n");
+    }
+
+    printf("y=%d\n",y);
+    */
+
+
+    x=lx+1;
+    if ( (wasIn = INSIDE(lx,y)) ) {
+      SET(lx,y);
+      lx--;
+      while(INSIDE(lx,y) && lx > 0) {
+	SET(lx,y);
+	lx--;
+      }
+    }
+
+    if (bxmin > lx) bxmin=lx;
+    
+    while(x <= xsize-1) {
+      /*  printf("x=%d\n",x); */
+      if (wasIn) {
+	
+	if (INSIDE(x,y)) {
+	  /* case 1: was inside, am still inside */
+	  SET(x,y);
+	} else {
+	  /* case 2: was inside, am no longer inside: just found the
+	     right edge of a span */
+	  ST_STACK(direction,dadLx,dadRx,lx,(x-1),y);
+
+	  if (bxmax < x) bxmax=x;
+
+	  wasIn=0;
+	}
+      } else {
+	if (x>rx) goto EXT;
+	if (INSIDE(x,y)) {
+	  SET(x,y);
+	  /* case 3: Wasn't inside, am now: just found the start of a new run */
+	  wasIn=1;
+	    lx=x;
+	} else {
+	  /* case 4: Wasn't inside, still isn't */
+	}
+      }
+      x++;
+    }
+  EXT: /* out of loop */
+    if (wasIn) {
+      /* hit an edge of the frame buffer while inside a run */
+      ST_STACK(direction,dadLx,dadRx,lx,(x-1),y);
+      if (bxmax < x) bxmax=x;
+    }
+  }
+  
+  /*   printf("lx=%d rx=%d dadLx=%d dadRx=%d y=%d direction=%d\n",lx,rx,dadLx,dadRx,y,direction); 
+       printf("bounding box: [%d,%d] - [%d,%d]\n",bxmin,bymin,bxmax,bymax); */
+
+  for(y=bymin;y<=bymax;y++) for(x=bxmin;x<=bxmax;x++) if (btm_test(btm,x,y)) i_ppix(im,x,y,dcol);
+
+  btm_destroy(btm);
+  mm_log((1, "DESTROY\n"));
+  llist_destroy(st);
+}
+
+static struct i_bitmap *
+i_flood_fill_low(i_img *im,int seedx,int seedy,
+                 int *bxminp, int *bxmaxp, int *byminp, int *bymaxp) {
+  int lx,rx;
+  int y;
+  int direction;
+  int dadLx,dadRx;
+
+  int wasIn=0;
+  int x=0;
+
+  /*  int tx,ty; */
+
+  int bxmin=seedx,bxmax=seedx,bymin=seedy,bymax=seedy;
+
+  struct llist *st;
+  struct i_bitmap *btm;
+
+  int channels,xsize,ysize;
+  i_color cval,val;
+
   channels=im->channels;
   xsize=im->xsize;
   ysize=im->ysize;
@@ -852,8 +1234,83 @@ i_flood_fill(i_img *im,int seedx,int seedy,i_color *dcol) {
   /*   printf("lx=%d rx=%d dadLx=%d dadRx=%d y=%d direction=%d\n",lx,rx,dadLx,dadRx,y,direction); 
        printf("bounding box: [%d,%d] - [%d,%d]\n",bxmin,bymin,bxmax,bymax); */
 
-  for(y=bymin;y<=bymax;y++) for(x=bxmin;x<=bxmax;x++) if (btm_test(btm,x,y)) i_ppix(im,x,y,dcol);
+  llist_destroy(st);
+
+  *bxminp = bxmin;
+  *bxmaxp = bxmax;
+  *byminp = bymin;
+  *bymaxp = bymax;
+
+  return btm;
+}
+
+void
+i_flood_cfill(i_img *im, int seedx, int seedy, i_fill_t *fill) {
+  int bxmin, bxmax, bymin, bymax;
+  struct i_bitmap *btm;
+  int x, y;
+  int start;
+
+  btm = i_flood_fill_low(im, seedx, seedy, &bxmin, &bxmax, &bymin, &bymax);
+
+  if (im->bits == i_8_bits && fill->fill_with_color) {
+    i_color *line = mymalloc(sizeof(i_color) * (bxmax - bxmin));
+    i_color *work = NULL;
+    if (fill->combine)
+      work = mymalloc(sizeof(i_color) * (bxmax - bxmin));
+
+    for(y=bymin;y<=bymax;y++) {
+      x = bxmin;
+      while (x < bxmax) {
+        while (x < bxmax && !btm_test(btm, x, y)) {
+          ++x;
+        }
+        if (btm_test(btm, x, y)) {
+          start = x;
+          while (x < bxmax && btm_test(btm, x, y)) {
+            ++x;
+          }
+          if (fill->combine)
+            i_glin(im, start, x, y, line);
+          (fill->fill_with_color)(fill, start, y, x-start, im->channels, 
+                                  line, work);
+          i_plin(im, start, x, y, line);
+        }
+      }
+    }
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
+  else {
+    i_fcolor *line = mymalloc(sizeof(i_fcolor) * (bxmax - bxmin));
+    i_fcolor *work = NULL;
+    if (fill->combinef)
+      work = mymalloc(sizeof(i_fcolor) * (bxmax - bxmin));
+    
+    for(y=bymin;y<=bymax;y++) {
+      x = bxmin;
+      while (x < bxmax) {
+        while (x < bxmax && !btm_test(btm, x, y)) {
+          ++x;
+        }
+        if (btm_test(btm, x, y)) {
+          start = x;
+          while (x < bxmax && btm_test(btm, x, y)) {
+            ++x;
+          }
+          if (fill->combinef)
+            i_glinf(im, start, x, y, line);
+          (fill->fill_with_fcolor)(fill, start, y, x-start, im->channels, 
+                                   line, work);
+          i_plinf(im, start, x, y, line);
+        }
+      }
+    }
+    myfree(line);
+    if (work)
+      myfree(work);
+  }
 
   btm_destroy(btm);
-  llist_destroy(st);
 }
