@@ -18,8 +18,16 @@ static char *i_format_list[]={
 #endif
 #ifdef HAVE_LIBTT
   "tt",
+#endif
+#ifdef HAVE_WIN32
+  "w32",
+#endif
+#ifdef HAVE_FT2
+  "ft2",
 #endif
   "raw",
   "pnm",
+  "bmp",
+  "tga",
   NULL};
 
