@@ -15,22 +15,23 @@
 #ifdef IMAGER_DEBUG_MALLOC
 
 #define mymalloc(x) (mymalloc_file_line((x), __FILE__, __LINE__))
+#define myrealloc(x,y) (myrealloc_file_line((x),(y), __FILE__, __LINE__))
 #define myfree(x) (myfree_file_line((x), __FILE__, __LINE__))
 
-void  malloc_state();
-void* mymalloc_file_line(int size,char* file,int line);
-void* mymalloc_comm(int size,char *comm);
-
-void  myfree_file_line(void *p, char*file, int line);
-
-void  bndcheck_all(void);
+void  malloc_state       (void);
+void* mymalloc_file_line (size_t size, char* file, int line);
+void  myfree_file_line   (void *p, char*file, int line);
+void* myrealloc_file_line(void *p, size_t newsize, char* file,int line);
+void* mymalloc_comm      (int size, char *comm);
+void  bndcheck_all       (void);
 
 #else
 
 #define malloc_comm(a,b) (mymalloc(a))
-void malloc_state();
+void  malloc_state(void);
 void* mymalloc(int size);
-void myfree(void *p);
+void  myfree(void *p);
+void* myrealloc(void *p, size_t newsize);
 
 #endif /* IMAGER_MALLOC_DEBUG */
 
@@ -39,10 +40,10 @@ void myfree(void *p);
 #undef min
 #undef max
 #endif
+
+/* XXX Shouldn't these go away? */
+
 int min(int a,int b);
 int max(int x,int y);
-int myread(int fd,void *buf,int len);
-int mywrite(int fd,void *buf,int len);
-void interleave(unsigned char *inbuffer,unsigned char *outbuffer,int rowsize,int channels);
 
 #endif
