@@ -1,9 +1,5 @@
 package Imager::Fill;
 use strict;
-use vars qw($VERSION);
-
-$VERSION = "1.010";
-
 # this needs to be kept in sync with the array of hatches in fills.c
 my @hatch_types =
   qw/check1x1 check2x2 check4x4 vline1 vline2 vline4
@@ -23,7 +19,47 @@ $combine_types{mult} = $combine_types{multiply};
 $combine_types{'sub'}  = $combine_types{subtract};
 $combine_types{sat}  = $combine_types{saturation};
 
-*_color = \&Imager::_color;
+# this function tries to DWIM for color parameters
+#  color objects are used as is
+#  simple scalars are simply treated as single parameters to Imager::Color->new
+#  hashrefs are treated as named argument lists to Imager::Color->new
+#  arrayrefs are treated as list arguments to Imager::Color->new iff any
+#    parameter is > 1
+#  other arrayrefs are treated as list arguments to Imager::Color::Float
+
+sub _color {
+  my $arg = shift;
+  my $result;
+
+  if (ref $arg) {
+    if (UNIVERSAL::isa($arg, "Imager::Color")
+        || UNIVERSAL::isa($arg, "Imager::Color::Float")) {
+      $result = $arg;
+    }
+    else {
+      if ($arg =~ /^HASH\(/) {
+        $result = Imager::Color->new(%$arg);
+      }
+      elsif ($arg =~ /^ARRAY\(/) {
+        if (grep $_ > 1, @$arg) {
+          $result = Imager::Color->new(@$arg);
+        }
+        else {
+          $result = Imager::Color::Float->new(@$arg);
+        }
+      }
+      else {
+        $Imager::ERRSTR = "Not a color";
+      }
+    }
+  }
+  else {
+    # assume Imager::Color::new knows how to handle it
+    $result = Imager::Color->new($arg);
+  }
+
+  return $result;
+}
 
 sub new {
   my ($class, %hsh) = @_;
@@ -108,26 +144,10 @@ sub new {
       }
     }
 
-    # check that the segments supplied is an array ref
-    unless (ref $hsh{segments} && $hsh{segments} =~ /ARRAY/) {
-      $Imager::ERRSTR =
-        "segments must be an array reference or Imager::Fountain object";
-      return;
-    }
-
-    # make sure the segments are specified with colors
-    my @segments;
-    for my $segment (@{$hsh{segments}}) {
-      my @new_segment = @$segment;
-
-      $_ = _color($_) or return for @new_segment[3,4];
-      push @segments, \@new_segment;
-    }
-
     $self->{fill} =
       Imager::i_new_fill_fount($hsh{xa}, $hsh{ya}, $hsh{xb}, $hsh{yb},
                   $hsh{ftype}, $hsh{repeat}, $hsh{combine}, $hsh{super_sample},
-                  $hsh{ssample_param}, \@segments);
+                  $hsh{ssample_param}, $hsh{segments});
   }
   elsif (defined $hsh{image}) {
     $hsh{xoff} ||= 0;
@@ -170,20 +190,8 @@ sub combines {
 
 =head1 DESCRIPTION 
 
-Creates fill objects for use by most filled area drawing functions.
-
-All fills are created with the new method.
-
-=over
-
-=item new
-
-  my $fill = Imager::Fill->new(...);
-
-The parameters depend on the type of fill being created.  See below
-for details.
-
-=back
+Creates fill objects for use by some drawing functions, currently just
+the Imager box() method.
 
 The currently available fills are:
 
@@ -197,7 +205,7 @@ solid
 
 hatch
 
-=item *
+=item
 
 fountain (similar to gradients in paint software)
 
@@ -250,13 +258,11 @@ absolute value taken this is alpha combined with the target.
 
 =item lighten
 
-The higher value is taken from each channel of the fill and target
-pixels, which is then alpha combined with the target.
+The higher value is taken from each channel of the fill and target pixels, which is then alpha combined with the target.
 
 =item darken
 
-The higher value is taken from each channel of the fill and target
-pixels, which is then alpha combined with the target.
+The higher value is taken from each channel of the fill and target pixels, which is then alpha combined with the target.
 
 =item hue
 
@@ -398,7 +404,7 @@ You can call Imager::Fill->hatches for a list of hatch names.
 
   my $fill = Imager::Fill->new(fountain=>$ftype, 
        xa=>$xa, ya=>$ya, xb=>$xb, yb=>$yb, 
-       segments=>$segments, repeat=>$repeat, combine=>$combine, 
+       segment=>$segments, repeat=>$repeat, combine=>$combine, 
        super_sample=>$super_sample, ssample_param=>$ssample_param);
 
 This fills the given region with a fountain fill.  This is exactly the
