@@ -1,10 +1,5 @@
 package Imager::Matrix2d;
 use strict;
-use vars qw($VERSION);
-use Scalar::Util qw(reftype looks_like_number);
-use Carp qw(croak);
-
-$VERSION = "1.0010";
 
 =head1 NAME
 
@@ -19,9 +14,6 @@ $VERSION = "1.0010";
   $m4 = Imager::Matrix2d->shear(x=>$sx, y=>$sy);
   $m5 = Imager::Matrix2d->reflect(axis=>$axis);
   $m6 = Imager::Matrix2d->scale(x=>$xratio, y=>$yratio);
-  $m8 = Imager::Matric2d->matrix($v11, $v12, $v13,
-                                 $v21, $v22, $v23,
-                                 $v31, $v32, $v33);
   $m6 = $m1 * $m2;
   $m7 = $m1 + $m2;
   use Imager::Matrix2d qw(:handy);
@@ -61,8 +53,7 @@ require 'Exporter.pm';
 use overload 
   '*' => \&_mult,
   '+' => \&_add,
-  '""'=>\&_string,
-  "eq" => \&_eq;
+  '""'=>\&_string;
 
 =item identity()
 
@@ -100,9 +91,7 @@ sub rotate {
     return undef;
   }
 
-  if ($opts{'x'} || $opts{'y'}) {
-    $opts{'x'} ||= 0;
-    $opts{'y'} ||= 0;
+  if ($opts{'x'} && $opts{'y'}) {
     return $class->translate('x'=>-$opts{'x'}, 'y'=>-$opts{'y'})
       * $class->rotate(radians=>$angle)
         * $class->translate('x'=>$opts{'x'}, 'y'=>$opts{'y'});
@@ -118,35 +107,24 @@ sub rotate {
 
 =item translate(x=>$dx, y=>$dy)
 
-=item translate(x=>$dx)
-
-=item translate(y=>$dy)
-
 Translates by the specify amounts.
 
 =cut
-
 sub translate {
   my ($class, %opts) = @_;
 
-  if (defined $opts{'x'} || defined $opts{'y'}) {
-    my $x = $opts{'x'} || 0;
-    my $y = $opts{'y'} || 0;
-    return bless [ 1, 0, $x,
-                   0, 1, $y,
+  if (defined $opts{'x'} && defined $opts{'y'}) {
+    return bless [ 1, 0, $opts{'x'},
+                   0, 1, $opts{'y'},
                    0, 0, 1 ], $class;
   }
-
-  $Imager::ERRSTR = 'x or y parameter required';
+  
+  $Imager::ERRSTR = 'x and y parameters required';
   return undef;
 }
 
 =item shear(x=>$sx, y=>$sy)
 
-=item shear(x=>$sx)
-
-=item shear(y=>$sy)
-
 Shear by the given amounts.
 
 =cut
@@ -213,7 +191,7 @@ sub reflect {
 
 Scales at the given ratios.
 
-You can also specify a center for the scaling with the C<cx> and C<cy>
+You can also specify a center for the scaling with the cx and cy
 parameters.
 
 =cut
@@ -241,24 +219,6 @@ sub scale {
   }
 }
 
-=item matrix($v11, $v12, $v13, $v21, $v22, $v23, $v31, $v32, $v33)
-
-Create a matrix with custom co-efficients.
-
-=cut
-
-sub matrix {
-  my ($class, @self) = @_;
-
-  if (@self == 9) {
-    return bless \@self, $class;
-  }
-  else {
-    $Imager::ERRSTR = "9 co-efficients required";
-    return;
-  }
-}
-
 =item _mult()
 
 Implements the overloaded '*' operator.  Internal use.
@@ -267,41 +227,28 @@ Currently both the left and right-hand sides of the operator must be
 an Imager::Matrix2d.
 
 =cut
-
 sub _mult {
   my ($left, $right, $order) = @_;
 
-  if (ref($right)) {
-    if (reftype($right) eq "ARRAY") {
-      @$right == 9
-	or croak "9 elements required in array ref";
-      if ($order) {
-	($left, $right) = ($right, $left);
-      }
-      my @result;
-      for my $i (0..2) {
-	for my $j (0..2) {
-	  my $accum = 0;
-	  for my $k (0..2) {
-	    $accum += $left->[3*$i + $k] * $right->[3*$k + $j];
-	  }
-	  $result[3*$i+$j] = $accum;
-	}
-      }
-      return bless \@result, __PACKAGE__;
+  if (ref($right) && UNIVERSAL::isa($right, __PACKAGE__)) {
+    if ($order) {
+      ($left, $right) = ($right, $left);
     }
-    else {
-      croak "multiply by array ref or number";
+    my @result;
+    for my $i (0..2) {
+      for my $j (0..2) {
+        my $accum = 0;
+        for my $k (0..2) {
+          $accum += $left->[3*$i + $k] * $right->[3*$k + $j];
+        }
+        $result[3*$i+$j] = $accum;
+      }
     }
-  }
-  elsif (defined $right && looks_like_number($right)) {
-    my @result = map $_ * $right, @$left;
-
     return bless \@result, __PACKAGE__;
   }
   else {
-    # something we don't handle
-    croak "multiply by array ref or number";
+    # presumably N * matrix or matrix * N
+    return undef; # for now
   }
 }
 
@@ -339,7 +286,6 @@ newline.
 I tried to make it fairly nicely formatted.  You might disagree :)
 
 =cut
-
 sub _string {
   my ($m) = @_;
 
@@ -375,21 +321,6 @@ sub _string {
   $out;
 }
 
-=item _eq
-
-Implement the overloaded equality operator.
-
-Provided for older perls that don't handle magic auto generation of eq
-from "".
-
-=cut
-
-sub _eq {
-  my ($left, $right) = @_;
-
-  return $left . "" eq $right . "";
-}
-
 =back
 
 The following functions are shortcuts to the various constructors.
@@ -412,8 +343,6 @@ You can import these methods with:
 
 =item m2d_reflect()
 
-=item m2d_scale()
-
 =back
 
 =cut
@@ -450,12 +379,12 @@ Tony Cook <tony@develop-help.com>
 
 =head1 BUGS
 
-Needs a way to invert a matrix.
+Needs a way to invert matrixes.
 
 =head1 SEE ALSO
 
 Imager(3), Imager::Font(3)
 
-http://imager.perl.org/
+http://www.eecs.umich.edu/~addi/perl/Imager/
 
 =cut
